
/* SEMsgConsumer_dpm
 * A plugin that feeds dpm with
 * app-level notifications.
 *
 *
 * by Fabrizio Furano, CERN, Oct 2010
 */


#include "SEMsgConsumer_dpm.hh"

#include "../SEMsgProtoDefs.hh"
#include "../SimpleDebug.hh"
#include <cms/MapMessage.h>


#include "dpns_api.h"
#include "stdlib.h"
#include "serrno.h"
#include "string.h"


using namespace cms;


// A few functions borrowed from DPM ... useful to convert an acl into the Cns format
gid_t cvt_group(char *p) {
  char *dp;
  gid_t gid;

  if (isdigit (*p)) {
    gid = strtol (p, &dp, 10);
    if (*dp != '\0') {
      Error("cvt_group", "Invalid group '" << p << "'");
      return (-1);
    }
  } else {
    if (strcmp (p, "root") == 0)
      gid = 0;
    else if (Cns_getgrpbynam (p, &gid) < 0) {
      Error("cvt_group", "Invalid group '" << p << "'");
      return (-1);
    }
  }
  return (gid);
}

mode_t cvt_perm(char *p) {
  char *dp;
  mode_t mode;
  char *q;

  if (isdigit (*p)) {
    mode = strtol (p, &dp, 8);
    if (*dp != '\0' || mode > 7) {
      Error("cvt_perm", "Invalid mode '" << p << "'");
      return (-1);
    }
  } else {
    if (strlen (p) > 3) {
      Error("cvt_perm", "Invalid mode '" << p << "'");
      return (-1);
    }
    mode = 0;
    q = p;
    while (*q) {
      switch (*q) {
      case 'r':
	mode |= 4;
	break;
      case 'w':
	mode |= 2;
	break;
      case 'x':
	mode |= 1;
      case '-':
	break;
      default:
	Error("cvt_perm", "Invalid mode '" << p << "'");
	return (-1);
      }
      q++;
    }
  }
  return (mode);
}

uid_t cvt_user(char *p) {
  char *dp;
  uid_t uid;
  
  if (isdigit (*p)) {
    uid = strtol (p, &dp, 10);
    if (*dp != '\0') {
      Error("cvt_user", "Invalid user '" << p << "'");  
      return (-1);
    }
  } else {
    if (strcmp (p, "root") == 0)
      uid = 0;
    else if (Cns_getusrbynam (p, &uid) < 0) {
      Error("cvt_user", "Invalid user '" << p << "'");  
      return (-1);
    }
  }
  return (uid);
}


int parse_entries2(char *entries, struct Cns_acl *acl) {
  char *entry = entries;
  int n;
  int nentries = 0;
  char *p;
  char *p_id;
  char *q;

  while (entry) {
    if ((p = strchr (entry, ','))) {
      *p = '\0';
      p++;
    }
    if ((q = strchr (entry, ':')) == NULL) {
      Error("parse_entries", "Missing colon delimiter in '" << entries << "'");
      return (-1);
    }
    *q = '\0';
    if (strcmp (entry, "d") == 0 || strcmp (entry, "default") == 0) {
      entry = q + 1;
      if ((q = strchr (entry, ':')) == NULL) {
	Error("parse_entries", "Missing colon delimiter (check2) in '" << entries << "'");
	return (-1);
      }
      *q = '\0';
      acl->a_type = CNS_ACL_DEFAULT;
    } else
      acl->a_type = 0;
    acl->a_id = 0;
    if (strcmp (entry, "u") == 0 || strcmp (entry, "user") == 0) {
      if (*(q + 1) == ':') {
	acl->a_type |= CNS_ACL_USER_OBJ;
	q++;
      } else
	acl->a_type |= CNS_ACL_USER;
    } else if (strcmp (entry, "g") == 0 || strcmp (entry, "group") == 0) {
      if (*(q + 1) == ':') {
	acl->a_type |= CNS_ACL_GROUP_OBJ;
	q++;
      } else
	acl->a_type |= CNS_ACL_GROUP;
    } else if (strcmp (entry, "m") == 0 || strcmp (entry, "mask") == 0) {
      acl->a_type |= CNS_ACL_MASK;
      if (*(q + 1) == ':')
	q++;
    } else if (strcmp (entry, "o") == 0 || strcmp (entry, "other") == 0) {
      acl->a_type |= CNS_ACL_OTHER;
      if (*(q + 1) == ':')
	q++;
    }
    q++;
    if (acl->a_type == CNS_ACL_USER ||
	acl->a_type == CNS_ACL_GROUP ||
	acl->a_type == (CNS_ACL_DEFAULT | CNS_ACL_USER) ||
	acl->a_type == (CNS_ACL_DEFAULT | CNS_ACL_GROUP)) {
      p_id = q;
      if ((q = strchr (q, ':')) == NULL) {
	  Error("parse_entries", "Missing colon delimiter (check3) in '" << entries << "'");
	  return (-1);
      } else
	*q = '\0';
      if (acl->a_type == CNS_ACL_USER ||
	  acl->a_type == (CNS_ACL_DEFAULT | CNS_ACL_USER)) {
	if ((acl->a_id = cvt_user (p_id)) < 0)
	  return (-1);
      } else {
	if ((acl->a_id = cvt_group (p_id)) < 0)
	  return (-1);
      }
      q++;
    }

    if ((n = cvt_perm (q)) < 0)
      return (-1);
    else
      acl->a_perm = n;

    acl++;
    nentries++;
    entry = p;
  }
  return (nentries);
}
int parse_entries(char *entries, struct Cns_acl *acl) {
  char *e = strdup(entries);
  int ret = parse_entries2(e, acl);
  free(e);
  return ret;
}


// ----------------------


SEMsgConsumer_dpm::SEMsgConsumer_dpm( const std::string& brokerURI,
				      const std::string& destname,
				      bool useTopic,
				      bool sessionTransacted ):
  SEMsgConsumer(brokerURI, destname, useTopic, sessionTransacted) {

  char *fname = "SEMsgConsumer_dpm::SEMsgConsumer_dpm";

  // The DPM client needs these envvars to work
  // We don't want to mess up with the startup scripts, hence we get them
  // from the config file
  std::string s;
  s = CFG->GetString("consd.dpmplugin.host", "");
  if (s.length() > 0)
       setenv("DPNS_HOST", s.c_str(), 1);
  Info(SimpleDebug::kLOW, fname, "DPNS_HOST: " << getenv("DPNS_HOST"));

     s = CFG->GetString("consd.cert", "");
     if (s.length() > 0)
       setenv("X509_USER_CERT", s.c_str(), 1);
     Info(SimpleDebug::kLOW, fname, "X509_USER_CERT: " << getenv("X509_USER_CERT"));

     s = CFG->GetString("consd.certkey", "");
     if (s.length() > 0)
       setenv("X509_USER_KEY", s.c_str(), 1);
     Info(SimpleDebug::kLOW, fname, "X509_USER_KEY: " << getenv("X509_USER_KEY"));

     setenv("CSEC_MECH", "ID", 1);
     Info(SimpleDebug::kLOW, fname, "CSEC_MECH: " << getenv("CSEC_MECH"));
   };






void SEMsgConsumer_dpm::onMessageChmod( const cms::MapMessage* m ) {
  const char *fname="SEMsgConsumer_dpm::onMessageChmod";

   std::string fn;
   long p = 0;
   int retval = 0;
   char **fqans = 0;
   int nfqans = 0;

   Info(SimpleDebug::kMEDIUM, fname, "Starting.");

   // Impersonate the sender of this message, if specified
   if (m->propertyExists(PROTO_AUTHDN)) {
     std::string authDN = m->getStringProperty(PROTO_AUTHDN);
     char DN[1024], voname[1024];

     Info(SimpleDebug::kHIGH, fname, "Using authDN:" << authDN);

     unpackAuthstuff((char *)authDN.c_str(), DN, voname, &fqans, &nfqans);

     Cns_client_setAuthorizationId (0, 0, "GSI", DN);
     Cns_client_setVOMS_data (voname, fqans, nfqans);
   }

   // Now invoke the lfc API with the info we got, and do whatever we need to do
   int cnt = 1;
   if (m->propertyExists(PROTO_NENTRIES))
     cnt = m->getIntProperty(PROTO_NENTRIES);
  
   if ( (cnt == 1) && (getStringField(fn, m, PROTO_CHMOD_FILENAME_S, -1, fname) > 0) && 
	(getLongField(p, m, PROTO_CHMOD_UNIXPERM_L, -1, fname) > 0)) {

     Info(SimpleDebug::kHIGH, fname, PROTO_CHMOD_FILENAME_S << "=" << fn <<
	  " " << PROTO_CHMOD_UNIXPERM_L << "=" << p);

     retval = dpns_chmod(fn.c_str(), p);

     if (retval)
       Error(fname, "dpns_chmod('" << fn << "', " << p << ") returned " << retval << ": " << strerror(serrno));

  }
    else {

      dpns_starttrans(getenv("DPM_HOST"), "Session for an incoming SEMsg bulk message");

      for (int i = 1; i < cnt+1; i++) {
	if (getStringField(fn, m, PROTO_CHMOD_FILENAME_S, i, fname) < 0) continue;

	if (getLongField(p, m, PROTO_CHMOD_UNIXPERM_L, i, fname) < 0) continue;

	retval = dpns_chmod(fn.c_str(), p);

	if (retval)
	  Error(fname, "dpns_chmod('" << fn << "', " << p << ") returned " << retval << ": " << strerror(serrno));
      }

    dpns_endtrans();
      
    }

   if (m->propertyExists(PROTO_AUTHDN)) {
    Cns_client_resetAuthorizationId();
    freeFqans(fqans, nfqans);
   }

}


void SEMsgConsumer_dpm::onMessageAddReplica( const cms::MapMessage* m ) {
  const char *fname="SEMsgConsumer_dpm::onMessageAddReplica";

  std::string lfn, guid, csumtype, csumvalue, server, sfn, status, f_type;
  long long size;
  long mode;
  char **fqans = 0;
  int nfqans = 0;

  int retval = 0;
  Info(SimpleDebug::kMEDIUM, fname, "Starting.");


  // Impersonate the sender of this message, if specified
  if (m->propertyExists(PROTO_AUTHDN)) {
    std::string authDN = m->getStringProperty(PROTO_AUTHDN);
    char DN[1024], voname[1024];
    
    Info(SimpleDebug::kHIGH, fname, "Using authDN:" << authDN);
    
    unpackAuthstuff((char *)authDN.c_str(), DN, voname, &fqans, &nfqans);
    
    Cns_client_setAuthorizationId (0, 0, "GSI", DN);
    Cns_client_setVOMS_data (voname, fqans, nfqans);
  }


  // Now invoke the lfc API with the info we got, and do whatever we need to do
  int cnt = 1;
  if (m->propertyExists(PROTO_NENTRIES))
    cnt = m->getIntProperty(PROTO_NENTRIES);
  
  if ( (cnt == 1) &&
       (getStringField(lfn,         m, PROTO_ADDREPLICA_LFN_S, -1, fname) > 0) && 
       (getStringField(guid,        m, PROTO_ADDREPLICA_GUID_S, -1, fname) >= 0) && 
       (getLongField(mode,          m, PROTO_ADDREPLICA_MODE_L, -1, fname) >= 0) &&
       (getLongLongField(size,      m, PROTO_ADDREPLICA_SIZE_LL, -1, fname) >= 0) &&
       (getStringField(csumtype,    m, PROTO_ADDREPLICA_CSUMTYPE_S, -1, fname) >= 0) &&
       (getStringField(csumvalue,   m, PROTO_ADDREPLICA_CSUMVALUE_S, -1, fname) >= 0) &&
       (getStringField(server,      m, PROTO_ADDREPLICA_SERVER_S, -1, fname) >= 0) && 
       (getStringField(server,      m, PROTO_ADDREPLICA_SERVER_S, -1, fname) >= 0) && 
       (getStringField(sfn,         m, PROTO_ADDREPLICA_SFN_S, -1, fname) > 0) ) {
    
    Info(SimpleDebug::kHIGH, fname,
	 PROTO_ADDREPLICA_LFN_S << "=" << lfn << " " <<
	 PROTO_ADDREPLICA_GUID_S << "=" << guid << " " <<
	 PROTO_ADDREPLICA_MODE_L << "=" << mode << " " <<
	 PROTO_ADDREPLICA_SIZE_LL << "=" << size << " " <<
	 PROTO_ADDREPLICA_CSUMTYPE_S << "=" << csumtype << " " <<
	 PROTO_ADDREPLICA_CSUMVALUE_S << "=" << csumvalue << " " <<
	 PROTO_ADDREPLICA_SERVER_S << "=" << server <<
	 PROTO_ADDREPLICA_SFN_S << "=" << sfn <<
	 PROTO_ADDREPLICA_STATUS_S << "=" << status <<
	 PROTO_ADDREPLICA_F_TYPE_S << "=" << f_type );
    
    struct dpns_fileid file_uniqueid;
    memset(&file_uniqueid, 0, sizeof(file_uniqueid));
    
    char st = '-';
    if (status.length() > 0) st = status[0];
    char ft = 'P';
    if (f_type.length() > 0) ft = f_type[0];

    retval = dpns_creatx(sfn.c_str(), mode, &file_uniqueid);
    if (retval)
      Error(fname,
	    "dpns_creatx(...) returned " << retval << ": " << strerror(serrno));
    retval = dpns_setfsize(sfn.c_str(), &file_uniqueid, size);
    if (retval)
      Error(fname,
	    "dpns_setfsize(...) returned " << retval << ": " << strerror(serrno));
    retval = dpns_addreplica(guid.c_str(), &file_uniqueid, server.c_str(), sfn.c_str(), st, ft, NULL, NULL);
    if (retval)
      Error(fname,
	    "dpns_addreplica(...) returned " << retval << ": " << strerror(serrno));


  }
    else {

      dpns_starttrans(getenv("DPM_HOST"), "Session for an incoming SEMsg bulk message");


      for (int i = 1; i < cnt+1; i++) {

	if (getStringField(lfn,         m, PROTO_ADDREPLICA_LFN_S, i, fname) <= 0) continue;
	if (getStringField(guid,        m, PROTO_ADDREPLICA_GUID_S, i, fname) < 0) continue;
	if (getLongField(mode,           m, PROTO_ADDREPLICA_MODE_L, i, fname) < 0) continue;
	if (getLongLongField(size,      m, PROTO_ADDREPLICA_SIZE_LL, i, fname) < 0) continue;
	if (getStringField(csumtype,    m, PROTO_ADDREPLICA_CSUMTYPE_S, i, fname) < 0) continue;
	if (getStringField(csumvalue,   m, PROTO_ADDREPLICA_CSUMVALUE_S, i, fname) < 0) continue;
	if (getStringField(server,      m, PROTO_ADDREPLICA_SERVER_S, i, fname) < 0) continue;
	if (getStringField(sfn,         m, PROTO_ADDREPLICA_SFN_S, i, fname) <= 0) continue;

	struct dpns_fileid file_uniqueid;
	memset(&file_uniqueid, 0, sizeof(file_uniqueid));

	char st = '-';
	if (status.length() > 0) st = status[0];
	char ft = 'P';
	if (f_type.length() > 0) ft = f_type[0];

	retval = dpns_creatx(sfn.c_str(), mode, &file_uniqueid);
	if (retval)
	  Error(fname,
		"dpns_creatx(...) returned " << retval << ": " << strerror(serrno));
	retval = dpns_setfsize(sfn.c_str(), &file_uniqueid, size);
	if (retval)
	  Error(fname,
		"dpns_setfsize(...) returned " << retval << ": " << strerror(serrno));
	retval = dpns_addreplica(guid.c_str(), &file_uniqueid,
				 server.c_str(), sfn.c_str(),
				 st, ft,
				 NULL, NULL);
	if (retval)
	  Error(fname,
		"dpns_addreplica(...) returned " << retval << ": " << strerror(serrno));

      }
      
    dpns_endtrans();

    } // else


  if (m->propertyExists(PROTO_AUTHDN)) {
    Cns_client_resetAuthorizationId();
    freeFqans(fqans, nfqans);
  }


}










void SEMsgConsumer_dpm::onMessageFileNotAvailable( const cms::MapMessage* m ) {
const char *fname="SEMsgConsumer_dpm::onMessageFileNotAvailable";
  std::string fn;
  int retval = 0;
  char **fqans = 0;
  int nfqans = 0;

  Info(SimpleDebug::kMEDIUM, fname, "Starting.");

  // Now invoke the lfc API with the info we got, and do whatever we need to do


  // Impersonate the sender of this message, if specified
  if (m->propertyExists(PROTO_AUTHDN)) {
    std::string authDN = m->getStringProperty(PROTO_AUTHDN);
    char DN[1024], voname[1024];
    
    Info(SimpleDebug::kHIGH, fname, "Using authDN:" << authDN);
    
    unpackAuthstuff((char *)authDN.c_str(), DN, voname, &fqans, &nfqans);
    
    Cns_client_setAuthorizationId (0, 0, "GSI", DN);
    Cns_client_setVOMS_data (voname, fqans, nfqans);
  }


  int cnt = 1;
  if (m->propertyExists(PROTO_NENTRIES))
    cnt = m->getIntProperty(PROTO_NENTRIES);
  
  Info(SimpleDebug::kHIGH, fname, "NEntries=" << cnt);
  
  if ( (cnt == 1) && (getStringField(fn, m, PROTO_FILENOTAVAILABLE_FILENAME_S, -1, fname) > 0) ) {

    retval = dpns_setrstatus(fn.c_str(), 'P');

    if (retval)
      Error(fname, "dpns_setrstatus('" << fn << "', 'P') returned " << retval << ": " << strerror(serrno));
  }
  else {

    dpns_starttrans(getenv("DPM_HOST"), "Session for an incoming SEMsg bulk message");
    
    for (int i = 1; i < cnt+1; i++) {
      if (getStringField(fn, m, PROTO_FILENOTAVAILABLE_FILENAME_S, i, fname) > 0) {
	
	retval = dpns_setrstatus(fn.c_str(), 'P');
	
	if (retval)
	  Error(fname, "dpns_setrstatus('" << fn << "', 'P') returned " << retval << ": " << strerror(serrno));
	
      }
    }

    dpns_endtrans();

    
  }
  

  if (m->propertyExists(PROTO_AUTHDN)) {
    Cns_client_resetAuthorizationId();
    freeFqans(fqans, nfqans);
  }
  
}










void SEMsgConsumer_dpm::onMessageSetACL( const cms::MapMessage* m ) {
  const char *fname="SEMsgConsumer_dpm::onMessageSetACL";

  std::string fn, stracl;
  struct dpns_acl acl[CA_MAXACLENTRIES];

  int retval = 0;
  int parseret = 0;

  char **fqans = 0;
  int nfqans = 0;

  Info(SimpleDebug::kMEDIUM, fname, "Starting.");

  // Now invoke the lfc API with the info we got, and do whatever we need to do
  
  // Impersonate the sender of this message, if specified
  if (m->propertyExists(PROTO_AUTHDN)) {
    std::string authDN = m->getStringProperty(PROTO_AUTHDN);
    char DN[1024], voname[1024];
    
    Info(SimpleDebug::kHIGH, fname, "Using authDN:" << authDN);
    
    unpackAuthstuff((char *)authDN.c_str(), DN, voname, &fqans, &nfqans);
    
    Cns_client_setAuthorizationId (0, 0, "GSI", DN);
    Cns_client_setVOMS_data (voname, fqans, nfqans);
  }


  int cnt = 1;
  if (m->propertyExists(PROTO_NENTRIES))
    cnt = m->getIntProperty(PROTO_NENTRIES);
  
  Info(SimpleDebug::kHIGH, fname, "NEntries=" << cnt);
  
  if ( (cnt == 1) && (getStringField(fn, m, PROTO_SETACL_FILENAME_S, -1, fname) > 0) ) {

    getStringField(stracl, m, PROTO_SETACL_ACLENTRIES_S, -1, fname);

    if ((parseret = parse_entries((char *)stracl.c_str(), acl)) > 0) {
      retval = dpns_setacl((char *)fn.c_str(), parseret, acl);
      if (retval)
	Error(fname, "dpns_setacl('" << fn << "', '" << stracl << "') returned " << retval << ": " << strerror(serrno));
    }
    else {
      Error(fname, "parse_entries('" << stracl << "') returned " << parseret);
      retval = -1;
    }


  }
  else {

    dpns_starttrans(getenv("DPM_HOST"), "Session for an incoming SEMsg bulk message");
    
    for (int i = 1; i < cnt+1; i++) {
      if (getStringField(fn, m, PROTO_SETACL_FILENAME_S, i, fname) > 0) {

	getStringField(stracl, m, PROTO_SETACL_ACLENTRIES_S, i, fname);

	if ((parseret = parse_entries((char *)stracl.c_str(), acl)) > 0) {
	  retval = dpns_setacl((char *)fn.c_str(), parseret, acl);
	  if (retval)
	    Error(fname, "dpns_setacl('" << fn << "', '" << stracl << "') returned " << retval << ": " << strerror(serrno));
	}
	else {
	  Error(fname, "parse_entries('" << stracl << "') returned " << parseret);
	  retval = -1;
	}

	if (retval)
	  Error(fname, "dpns_setacl('" << fn << "', " << stracl << "') returned " << retval << ": " << strerror(serrno));
	
      }
    }

    dpns_endtrans();

    
  }

  if (m->propertyExists(PROTO_AUTHDN)) {
    Cns_client_resetAuthorizationId();
    freeFqans(fqans, nfqans);
  }

  
}








//----------------------------
// The hook function to get an instance of the class
//  when the plugin is loaded
//----------------------------

extern "C" SEMsgConsumer *GetSEMsgConsumer(GetSEMsgConsumerArgs)
{
   return (SEMsgConsumer *)new SEMsgConsumer_dpm(brokerURI, destname, useTopic, sessionTransacted);
}
