
/* SEMsgConsumer_lfc
 * A plugin that feeds lfc with
 * app-level notifications.
 *
 *
 * by Fabrizio Furano, CERN, Nov 2010
 */


#include "SEMsgConsumer_lfc.hh"

#include "../SEMsgProtoDefs.hh"
#include "../SimpleDebug.hh"
#include <cms/MapMessage.h>


#include "lfc_api.h"
#include "serrno.h"

#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

using namespace cms;

// Implement a mkdir -p
// We would like this procssing to start from the last token, not the first one
// so that it is optimized for directory paths that already exist almost completely
// avoiding the unnecessary iterations

int mkdirminusp(const char *path) {
  int c;
  char pathbuf[1024];
  char *p = NULL, *startp = NULL, *endp = NULL;
  mode_t mask, mode;
  int mflag = 0;
  struct lfc_filestat statbuf;
  
  if (lfc_lstat (pathbuf, &statbuf) == 0) {
    if ((statbuf.filemode & S_IFDIR) == 0) {
      return EEXIST;
    }
  }

  strncpy(pathbuf, path, sizeof(pathbuf));
  pathbuf[sizeof(pathbuf)-1] = '\0';

  if ((startp = strstr (pathbuf, ":/")) == NULL)
    startp = pathbuf;
  else
    startp++;

  endp = strrchr (pathbuf, '/');
  p = endp;
  while (p > startp) {
    *p = '\0';
    if (lfc_access (pathbuf, F_OK) == 0) break;
    p = strrchr (pathbuf, '/');
  }

  /* make sure that mask allow write/execute for owner */
  if (mask & 0300)
    lfc_umask (mask & ~ 0300);

  while (p <= endp) {
    *p = '/';
    if (p == endp && mask & 0300)
      lfc_umask (mask);
    c = lfc_mkdir (pathbuf, 0777);

    if (c < 0 && serrno != EEXIST) {
      return(serrno);
    }

    p += strlen (p);
  }

  if (c == 0      /* directory successfully created */
      && mflag) {
    /* must set requested mode + possible S_ISGID */
    if (lfc_stat (pathbuf, &statbuf) < 0) {
      return(serrno);
    }

    if (lfc_chmod (pathbuf, (mode | statbuf.filemode & S_ISGID)) < 0) {
      return(serrno);
    }
  }


  return 0;
}



SEMsgConsumer_lfc::SEMsgConsumer_lfc( const std::string& brokerURI,
                                      const std::string& destname,
                                      bool useTopic,
                                      bool sessionTransacted ):
  SEMsgConsumer(brokerURI, destname, useTopic, sessionTransacted) {

  char *fname = "SEMsgConsumer_lfc::SEMsgConsumer_lfc";

  // The LFC client needs these envvars to work
  // We don't want to mess up with the startup scripts, hence we get them
  // from the config file
  std::string s;
  s = CFG->GetString("consd.lfcplugin.host", "");
  if (s.length() > 0)
       setenv("LFC_HOST", s.c_str(), 1);
  Info(SimpleDebug::kLOW, fname, "LFC_HOST: " << getenv("LFC_HOST"));

     s = CFG->GetString("consd.cert", "");
     if (s.length() > 0)
       setenv("X509_USER_CERT", s.c_str(), 1);
     Info(SimpleDebug::kLOW, fname, "X509_USER_CERT: " << getenv("X509_USER_CERT"));

     s = CFG->GetString("consd.certkey", "");
     if (s.length() > 0)
       setenv("X509_USER_KEY", s.c_str(), 1);
     Info(SimpleDebug::kLOW, fname, "X509_USER_KEY: " << getenv("X509_USER_KEY"));

     setenv("CSEC_MECH", "ID", 1);
     Info(SimpleDebug::kLOW, fname, "CSEC_MECH: " << getenv("CSEC_MECH"));
   };



void SEMsgConsumer_lfc::onMessageFileNotAvailable( const cms::MapMessage* m ) {
const char *fname="SEMsgConsumer_lfc::onMessageFileNotAvailable";
  std::string fn;
  int retval = 0;
  char **fqans = 0;
  int nfqans = 0;

  Info(SimpleDebug::kMEDIUM, fname, "Starting.");

  // Impersonate the sender of this message, if specified
  if (m->propertyExists(PROTO_AUTHDN)) {
    std::string authDN = m->getStringProperty(PROTO_AUTHDN);
    char DN[1024], voname[1024];

    Info(SimpleDebug::kHIGH, fname, "Using authDN:" << authDN);

    unpackAuthstuff((char *)authDN.c_str(), DN, voname, &fqans, &nfqans);

    Cns_client_setAuthorizationId (0, 0, "GSI", DN);
    Cns_client_setVOMS_data (voname, fqans, nfqans);
  }

  // Now invoke the lfc API with the info we got, and do whatever we need to do

  int cnt = 1;
  if (m->propertyExists(PROTO_NENTRIES))
    cnt = m->getIntProperty(PROTO_NENTRIES);

  Info(SimpleDebug::kHIGH, fname, "NEntries=" << cnt);

  if ( (cnt == 1) && (getStringField(fn, m, PROTO_FILENOTAVAILABLE_FILENAME_S, -1, fname) > 0) ) {

    retval = lfc_setrstatus(fn.c_str(), 'P');

    if (retval)
      Error(fname, "lfc_setrstatus('" << fn << "', 'P') returned " << retval << ": " << strerror(serrno));
  }
  else {

    lfc_starttrans(getenv("LFC_HOST"), "Session for an incoming SEMsg bulk message");

    for (int i = 1; i < cnt+1; i++) {
      if (getStringField(fn, m, PROTO_FILENOTAVAILABLE_FILENAME_S, i, fname) > 0) {

        retval = lfc_setrstatus(fn.c_str(), 'P');

        if (retval)
          Error(fname, "lfc_setrstatus('" << fn << "', 'P') returned " << retval << ": " << strerror(serrno));

      }

    }

    lfc_endtrans();

  }

  if (m->propertyExists(PROTO_AUTHDN)) {
    Cns_client_resetAuthorizationId();
    freeFqans(fqans, nfqans);
  }


}


void SEMsgConsumer_lfc::onMessageFileDeleted( const cms::MapMessage* m ) {
const char *fname="SEMsgConsumer_lfc::onMessageFileDeleted";
  std::string fn;
  int retval = 0;
  char **fqans = 0;
  int nfqans = 0;

  Info(SimpleDebug::kMEDIUM, fname, "Starting.");

  // Now invoke the lfc API with the info we got, and do whatever we need to do


  // Impersonate the sender of this message, if specified
  if (m->propertyExists(PROTO_AUTHDN)) {
    std::string authDN = m->getStringProperty(PROTO_AUTHDN);
    char DN[1024], voname[1024];

    Info(SimpleDebug::kHIGH, fname, "Using authDN:" << authDN);

    unpackAuthstuff((char *)authDN.c_str(), DN, voname, &fqans, &nfqans);

    Cns_client_setAuthorizationId (0, 0, "GSI", DN);
    Cns_client_setVOMS_data (voname, fqans, nfqans);
  }

  int cnt = 1;
  if (m->propertyExists(PROTO_NENTRIES))
    cnt = m->getIntProperty(PROTO_NENTRIES);

  Info(SimpleDebug::kHIGH, fname, "NEntries=" << cnt);

  if ( (cnt == 1) && (getStringField(fn, m, PROTO_FILEDELETED_FILENAME_S, -1, fname) > 0) ) {

    retval = lfc_setrstatus(fn.c_str(), 'D');

    if (retval)
      Error(fname, "lfc_setrstatus('" << fn << "', 'D') returned " << retval << ": " << strerror(serrno));
  }
  else {

    lfc_starttrans(getenv("LFC_HOST"), "Session for an incoming SEMsg bulk message");

    for (int i = 1; i < cnt+1; i++) {
      if (getStringField(fn, m, PROTO_FILEDELETED_FILENAME_S, i, fname) > 0) {

        retval = lfc_setrstatus(fn.c_str(), 'D');

        if (retval)
          Error(fname, "lfc_setrstatus('" << fn << "', 'D') returned " << retval << ": " << strerror(serrno));

      }

    }

    lfc_endtrans();

  }

  if (m->propertyExists(PROTO_AUTHDN)) {
    Cns_client_resetAuthorizationId();
    freeFqans(fqans, nfqans);
  }


}



void SEMsgConsumer_lfc::onMessageFileAvailable( const cms::MapMessage* m ) {
const char *fname="SEMsgConsumer_lfc::onMessageFileAvailable";
  std::string fn;
  int retval = 0;
  char **fqans = 0;
  int nfqans = 0;

  Info(SimpleDebug::kMEDIUM, fname, "Starting.");

  // Now invoke the lfc API with the info we got, and do whatever we need to do

  // Impersonate the sender of this message, if specified
  if (m->propertyExists(PROTO_AUTHDN)) {
    std::string authDN = m->getStringProperty(PROTO_AUTHDN);
    char DN[1024], voname[1024];

    Info(SimpleDebug::kHIGH, fname, "Using authDN:" << authDN);

    unpackAuthstuff((char *)authDN.c_str(), DN, voname, &fqans, &nfqans);

    Cns_client_setAuthorizationId (0, 0, "GSI", DN);
    Cns_client_setVOMS_data (voname, fqans, nfqans);
  }

  int cnt = 1;
  if (m->propertyExists(PROTO_NENTRIES))
    cnt = m->getIntProperty(PROTO_NENTRIES);
  
  Info(SimpleDebug::kHIGH, fname, "NEntries=" << cnt);
  
  if ( (cnt == 1) && (getStringField(fn, m, PROTO_FILEAVAILABLE_FILENAME_S, -1, fname) > 0) ) {

    retval = lfc_setrstatus(fn.c_str(), '-');

    if (retval)
      Error(fname, "lfc_setrstatus('" << fn << "', '-') returned " << retval << ": " << strerror(serrno));
  }
  else {
    
    lfc_starttrans(getenv("LFC_HOST"), "Session for an incoming SEMsg bulk message");
   
    for (int i = 1; i < cnt+1; i++) {
      if (getStringField(fn, m, PROTO_FILEAVAILABLE_FILENAME_S, i, fname) > 0) {
	
	retval = lfc_setrstatus(fn.c_str(), '-');
	
	if (retval)
	  Error(fname, "lfc_setrstatus('" << fn << "', '-') returned " << retval << ": " << strerror(serrno));
	
      }

    }

    lfc_endtrans();
    
  }
  
  if (m->propertyExists(PROTO_AUTHDN)) {
    Cns_client_resetAuthorizationId();
    freeFqans(fqans, nfqans);
  }

  
}



void SEMsgConsumer_lfc::onMessageAddReplica( const cms::MapMessage* m ) {
  const char *fname="SEMsgConsumer_lfc::onMessageAddReplica";

  std::string lfn, guid, csumtype, csumvalue, server, sfn;
  long long size;
  long mode;

  int retval = 0;
  char **fqans = 0;
  int nfqans = 0;

  Info(SimpleDebug::kMEDIUM, fname, "Starting.");


  // Now invoke the lfc API with the info we got, and do whatever we need to do

  // Impersonate the sender of this message, if specified
  if (m->propertyExists(PROTO_AUTHDN)) {
    std::string authDN = m->getStringProperty(PROTO_AUTHDN);
    char DN[1024], voname[1024];

    Info(SimpleDebug::kHIGH, fname, "Using authDN:" << authDN);

    unpackAuthstuff((char *)authDN.c_str(), DN, voname, &fqans, &nfqans);

    Cns_client_setAuthorizationId (0, 0, "GSI", DN);
    Cns_client_setVOMS_data (voname, fqans, nfqans);
  }

  int cnt = 1;
  if (m->propertyExists(PROTO_NENTRIES))
    cnt = m->getIntProperty(PROTO_NENTRIES);
  
  if ( (cnt == 1) &&
       (getStringField(lfn,         m, PROTO_ADDREPLICA_LFN_S, -1, fname) > 0) && 
       (getStringField(guid,        m, PROTO_ADDREPLICA_GUID_S, -1, fname) >= 0) && 
       (getLongField(mode,          m, PROTO_ADDREPLICA_MODE_L, -1, fname) >= 0) &&
       (getLongLongField(size,      m, PROTO_ADDREPLICA_SIZE_LL, -1, fname) >= 0) &&
       (getStringField(csumtype,    m, PROTO_ADDREPLICA_CSUMTYPE_S, -1, fname) >= 0) &&
       (getStringField(csumvalue,   m, PROTO_ADDREPLICA_CSUMVALUE_S, -1, fname) >= 0) &&
       (getStringField(server,      m, PROTO_ADDREPLICA_SERVER_S, -1, fname) >= 0) && 
       (getStringField(sfn,         m, PROTO_ADDREPLICA_SFN_S, -1, fname) > 0) ) {

    Info(SimpleDebug::kHIGH, fname,
	 PROTO_ADDREPLICA_LFN_S << "=" << lfn << " " <<
	 PROTO_ADDREPLICA_GUID_S << "=" << guid << " " <<
	 PROTO_ADDREPLICA_MODE_L << "=" << mode << " " <<
	 PROTO_ADDREPLICA_SIZE_LL << "=" << size << " " <<
	 PROTO_ADDREPLICA_CSUMTYPE_S << "=" << csumtype << " " <<
	 PROTO_ADDREPLICA_CSUMVALUE_S << "=" << csumvalue << " " <<
	 PROTO_ADDREPLICA_SERVER_S << "=" << server <<
	 PROTO_ADDREPLICA_SFN_S << "=" << sfn );

    struct lfc_filereg fr;

    fr.lfn = (char *)lfn.c_str();
    fr.guid = (char *)guid.c_str();
    fr.mode = mode;
    fr.size = size;
    fr.csumtype = (char *)csumtype.c_str();
    fr.csumvalue = (char *)csumvalue.c_str();
    fr.server = (char *)server.c_str();
    fr.sfn = (char *)sfn.c_str();

    // Make sure that the directory of the lfn exists
    if (lfn != "") {
      // Get this dir
      size_t p1 = lfn.rfind("/");
      std::string d = lfn.substr(0, p1);
      mkdirminusp(d.c_str());
    }

    int nbstatuses = 0;
    int *statuses = NULL;

    retval = lfc_registerfiles(1, &fr, &nbstatuses, &statuses);

    if (retval)
       Error(fname,
	     "lfc_registerfiles(...) returned " << retval << ": " << strerror(serrno));
    if (nbstatuses)
      Error(fname,
	    "lfc_registerfiles(...) returned " << *statuses << ": " << strerror(*statuses) << " for item 0");

    if (statuses) free(statuses);

  }
    else {
      struct lfc_filereg *fr = new lfc_filereg[cnt];
      int fr_cnt = 0;

      lfc_starttrans(NULL, "Session for an incoming SEMsg bulk message");


      for (int i = 1; i < cnt+1; i++) {
     
	if (getStringField(lfn,         m, PROTO_ADDREPLICA_LFN_S, i, fname) <= 0) continue;
	if (getStringField(guid,        m, PROTO_ADDREPLICA_GUID_S, i, fname) < 0) continue;
	if (getLongField(mode,          m, PROTO_ADDREPLICA_MODE_L, i, fname) < 0) continue;
	if (getLongLongField(size,      m, PROTO_ADDREPLICA_SIZE_LL, i, fname) < 0) continue;
	if (getStringField(csumtype,    m, PROTO_ADDREPLICA_CSUMTYPE_S, i, fname) < 0) continue;
	if (getStringField(csumvalue,   m, PROTO_ADDREPLICA_CSUMVALUE_S, i, fname) < 0) continue;
	if (getStringField(server,      m, PROTO_ADDREPLICA_SERVER_S, i, fname) < 0) continue;
	if (getStringField(sfn,         m, PROTO_ADDREPLICA_SFN_S, i, fname) <= 0) continue;

	fr[fr_cnt].lfn = strdup(lfn.c_str());
	fr[fr_cnt].guid = strdup(guid.c_str());
	fr[fr_cnt].mode = mode;
	fr[fr_cnt].size = size;
	fr[fr_cnt].csumtype = strdup(csumtype.c_str());
	fr[fr_cnt].csumvalue = strdup(csumvalue.c_str());
	fr[fr_cnt].server = strdup(server.c_str());
	fr[fr_cnt].sfn = strdup(sfn.c_str());

	// Make sure that the directory of the lfn exists
	if (lfn != "") {
	  // Get this dir
	  size_t p1 = lfn.rfind("/");
	  std::string d = lfn.substr(0, p1);
	  mkdirminusp(d.c_str());
	}

	fr_cnt++;
      }

      int nbstatuses;
      int *statuses;

      retval = lfc_registerfiles(fr_cnt, fr, &nbstatuses, &statuses);

      if (retval)
	Error(fname,
	      "lfc_registerfiles(" << fr_cnt << " ...) returned " << retval << ": " << strerror(serrno));

      for (int i = 0; i < nbstatuses; i++) {
	if (statuses[i])
	  Error(fname,
		"lfc_registerfiles(" << fr_cnt << " ...) returned " << statuses[i] << ": " << strerror(statuses[i]) <<
		" for item " << i);

      }


      // And now free all the silly strings that we had to allocate
      for (int i = 0; i < fr_cnt; i++) {
	if (fr[i].lfn) free(fr[i].lfn);
	if (fr[i].guid) free(fr[i].guid);
	if (fr[i].csumtype) free(fr[i].csumtype);
	if (fr[i].csumvalue) free(fr[i].csumvalue);
	if (fr[i].server) free(fr[i].server);
	if (fr[i].sfn) free(fr[i].sfn);
      }
      delete [] fr;


    lfc_endtrans();
      
    } // else

  if (m->propertyExists(PROTO_AUTHDN)) {
    Cns_client_resetAuthorizationId();
    freeFqans(fqans, nfqans);
  }


}









//----------------------------
// The hook function to get an instance of the class
//  when the plugin is loaded
//----------------------------

extern "C" SEMsgConsumer *GetSEMsgConsumer(GetSEMsgConsumerArgs)
{
   return (SEMsgConsumer *)new SEMsgConsumer_lfc(brokerURI, destname, useTopic, sessionTransacted);
}
